__title__ = ''
__package_name__ = 'mindsdb_dummy_data_handler'
__version__ = '0.0.1'
__description__ = 'Handler for tests that use data handlers'
__author__ = ''
__github__ = ''
__pypi__ = ''
__license__ = ''
__copyright__ = ''
